/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * different classes for different IN channels
 * they are instantiated and used by the Stream_mixer class (jmixer.cpp)
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <jutils.h>

#include "inchannels.h"

/* ----- Parent Class Channel ----- */

typedef void* (kickoff)(void*);

Channel::Channel() {
  volume = 1.0;
  speed = 100;
  time.h = time.m = time.s = 0;
  position = time.f = 0.0;
  playmode = PLAY;
  opened = false;
  on = false;
  update = false;
  running = false;
  seekable = false;
  rewinding = false;
  _thread_initialized = false;
}

void Channel::_thread_init() {
  if(_thread_initialized) return;

  func("Channel::thread_init()");
  if(pthread_mutex_init (&_mutex,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_cond_init (&_cond, NULL) == -1)
    error("error initializing POSIX thread condition"); 
  if(pthread_attr_init (&_attr) == -1)
    error("error initializing POSIX thread attribute");
  
  /* set the thread as detached
     see: man pthread_attr_init(3) */
  pthread_attr_setdetachstate(&_attr,PTHREAD_CREATE_DETACHED);

  _thread_initialized = true;
}

void Channel::_thread_destroy() {
  if(!_thread_initialized) return;
  opened = false;  

  /* we signal and then we check the thread
     exited by locking the conditional */
  if(running) {
    signal();
    lock(); unlock();
  }

  if(pthread_mutex_destroy(&_mutex) == -1)
    error("error destroying POSIX thread mutex");
  if(pthread_cond_destroy(&_cond) == -1)
    error("error destroying POSIX thread condition");
  if(pthread_attr_destroy(&_attr) == -1)
    error("error destroying POSIX thread attribute");
  _thread_initialized = false;
}

void Channel::run() {
  func("Channel::run() waits to lock");
  lock();
  func("Channel::run() locked");
  running = true;
  while(opened) {
    _res = _get_audio();
    this->wait();
  }
  running = false;
  unlock();
}

float Channel::mix(int *mixpcm) {
  if(_res>=1.0) return(_res);
  lock();
  in_smp = (*mixxx)(mixpcm,(short int *)buffo,frame,volume);
  unlock();
  return(_res);
}

bool Channel::set_mixer() {
  if(samplerate==44100) /* stereo */
    if(channels==2)
      mixxx = mixxx_stereo_44;
    else
      mixxx = mixxx_mono_44;
  else if(samplerate==22050)
    if(channels==2)
      mixxx = mixxx_stereo_22;
    else
      mixxx = mixxx_mono_22;
  else if(samplerate==16000)
    if(channels==2)
      mixxx = mixxx_stereo_16;
    else
      mixxx = mixxx_mono_16;
  else {
    warning("can't play a sound to %uKhz",samplerate);
    return(false);
  }
  return(true);
}

float Channel::upd_time() {
  long int secs;
  float res;

  update = false;

  res = (float)framepos/(float)frametot;

  if((res-time.f)>UPDATE_INTERVAL) {
    time.f = res;
    secs = framepos/fps;
    
    if(secs>3600) {
      time.h = (int) secs / 3600;
      secs -= time.h*3600;
    }
    if(secs>60) {
      time.m = (int) secs / 60;
      secs -= time.m*60;
    }
    time.s = (int) secs;
    update = true;
  }
  return(res);
}

#ifdef HAVE_VORBIS
/* ----- OggVorbis Channel ----- */

OggChannel::OggChannel() : Channel() {
  type = OGGCHAN;

  old_section = current_section = -1;

  oggfile = NULL;
}

OggChannel::~OggChannel() {
  if(on) stop();
  //  destroy();
  if(opened) clean();
  ov_clear(&vf);
}

void OggChannel::clean() {
  fclose(oggfile);
}

float OggChannel::_get_audio() {
  int res;
  float fres;

  old_section = current_section;  
  
  if(seekable)
    /* we check the position here 'cause ov_pcm_tell returns the NEXT frame */
    framepos = ov_pcm_tell(&vf);

  res = 
    ov_read(&vf, _buffone, 4096, 0, 2, 1, &current_section);

  //  func("readed %u",res);
  
  if(res<0) {
    warning("OggVorbis: bitstream error");
    in_smp = 0;
    return(3.0);
  }

  /*  if((res==0)||(old_section != current_section && old_section != -1)) {
     with this we should check out entering into a new logical bitstream
     left here, will manage later, maybe, one day, oh yes, should do */

  frame = res;
  buffo = (unsigned char *)_buffone;

  if(res==0) return(2.0);

  if(seekable)
    fres = upd_time();
  else fres = 0.0;

  //  func("returning %f",fres);

  return(fres);
}

int OggChannel::set(char *file) {
  int res = 0;

  lock();
  on = false;

  if(opened)
    clean();

  oggfile = fopen(file,"rb");
  if(oggfile==NULL) {
    warning("can't read file %s",file);
    unlock();
    return(res);
  } else func("opened %s",file);

  if(ov_open(oggfile,&vf,NULL,0) < 0) {
    warning("OggVorbis: input not a valid Ogg Vorbis audio stream.");
    fclose(oggfile);
    unlock();
    return(res);
  }

  vc = ov_comment(&vf, -1);
  vi = ov_info(&vf, -1);

  samplerate = vi->rate;
  channels = vi->channels;
  seekable = ov_seekable(&vf);

  if(!set_mixer()) {
    clean();
    unlock();
    return(res);
  } 

  /* pcm position */
  framepos = 0;
  /* that's a bit shamanic but works fine for me */
  fps = samplerate;
  /* time position */
  time.h = time.m = time.s = 0;
  /* floating point position */
  position = time.f = 0.0;

  /* check if seekable */
  res = (ov_seekable(&vf)) ? 1 : 2;
  seekable = (res>1) ? false : true;

  if(seekable)
    frametot = ov_pcm_total(&vf,-1);

  opened = true;
  
  unlock();
  
  if(!running)
    this->start();
  
  return(res);
}

bool OggChannel::play() {
  if(!on) {
    if(opened) {
      lock();
      on = true;
      if(time.f!=position) {

	/* this is equivalent to pos(position); */
	if(position==0.0) {
	  if(ov_pcm_seek(&vf,1)!=0) {
	    error("error in seeking channel for playing");
	    unlock();
	    return(false);
	  }
	} else {
	  if(ov_pcm_seek_page(&vf,(ogg_int64_t)((double)frametot * position))!=0) {
	    error("error in seeking channel for playing");
	    unlock();
	    return(false);
	  }
	}
	time.f = position;
	/* ----------------------------------- */

      }
      unlock();
    }
  }
  return(on);
}

bool OggChannel::stop() {
  lock();
  on = false;
  position = 0.0;
  unlock();
  return(!on);
}

bool OggChannel::pos(float pos) {

  if(on) {
    lock();

    if(pos==0.0) {
      if(ov_pcm_seek(&vf,1)!=0)
	return(false);
    } else {
      if(ov_pcm_seek_page(&vf,(ogg_int64_t)((double)frametot * pos))!=0)
	return(false);
    }
    time.f = position = pos;

    unlock();
  } else {
    position = pos;
  }

  return(true);
}

#endif

/* ----- Mpeg Channel ----- */

MpegChannel::MpegChannel() : Channel() {

  type = MP3CHAN;

  loader = NULL;
  server = NULL;
}

MpegChannel::~MpegChannel() {
  if(on) stop();
  //  destroy();
  if(opened) clean();
}

/* cleans decoder memory but does'nt kills the thread */
void MpegChannel::clean() {
  if(loader!=NULL) {
    delete loader;
    loader = NULL;
  }

  if(server!=NULL) {
    delete server;
    server = NULL;
  }
}

int MpegChannel::set(char *file) {
  int res = 0;

  lock();
  func("MpegChannel::set(%s)",file);

  on = false;
  
  if(opened)
    clean();
  
  loader = Soundinputstream::hopen(file,&res);
  if(loader == NULL) {
    warning("i don't see any stream flowing from %s",file);
    clean();
    unlock();
    return(0);
  }

  server = new Mpegtoraw(loader);
  if(server == NULL) {
    warning("MpegChannel::set(%s) - uuops! NULL Mpegtoraw",file);
    clean();
    unlock();
    return(0);
  }

  server->initialize(file);

  samplerate = server->getfrequency();
  channels = (server->getmode()<3) ? 2 : 1;

  bitrate = server->getbitrate();

  if(!set_mixer()) {
    func("MpegChannel::set(%s) - argh! set_mixer() reported failure",file);
    clean();
    unlock();
    return(0);
  }

  /* pcm position */
  framepos = 0;
  /* time position */
  time.h = time.m = time.s = 0;
  /* floating point position */
  position = time.f = 0.0;

  /* check if seekable */
  res = (loader->seekable) ? 1: 2;
  seekable = (loader->seekable) ? true : false;

  if(seekable) { /* setup position tracking variables */
    /* total chunks into bitstream (if seekable) */
    frametot = server->gettotalframe();
    /* size of a chunk in bytes */
    framesize = server->getpcmperframe()<<2;
    /* how much frames make second */
    fps = samplerate/server->getpcmperframe();
  } else framesize = (1152*channels)<<1; /* hardcoded mp3 framesize in bytes */

  opened = true;

  func("MpegChannel::set bitrate[%u] channels[%u]%sseekable",
       bitrate,channels,(seekable)?" ":" non ");
  
  unlock();

  if(!running)
    this->start();

  return(res);
}

float MpegChannel::_get_audio() {
  float res;
  /*  uhm, i guess that's not needed
      if(!server->run(-1)) // Initialize MPEG Layer 3
      return(3.0);
  */
  if(!server->run(1)) {
    func("in MpegChannel::_get_audio() server->run(1) returned 2.0");
    return(2.0);
    /* should i set errorcodes here? */
  }

  frame = framesize;

  if(seekable) {
    framepos = server->getcurrentframe();
    res = upd_time();
  } else res = 0.0;

  /* resampling deactivated (speed)
     if you are wondering: this is not enough for it.
     in_smp = resample((short int *)&server->rawdata[0],in_smp,buffo,speed);
     fprintf(stderr,"Mpeg: _get_audio : in_smp %u\n",in_smp);
  */
  buffo = (unsigned char *)server->rawdata;

  return(res);
}

bool MpegChannel::play() {
  lock();
  on = true;
  if(time.f!=position) {
    /* this is equivalent to pos() call */
    if(position==0.0) {
      server->setframe(0);
      _res = 0.0;
    } else {
      server->setframe((int) (frametot*position));
    }
    time.f = position;
  }
  unlock();
  return(on);
}

bool MpegChannel::stop() {
  lock();
  on = false;
  position = 0.0;
  unlock();
  return(true);
}

bool MpegChannel::pos(float pos) {
  if(on) {
    lock();

    if(pos==0.0)
      server->setframe(0);
    else
      server->setframe((int) (frametot*pos));

    time.f = position = pos;
    unlock();
  } else
    position = pos;
  
  return(true);
}

/* ----- LiveIN DSP Channel ----- */

LiveIn::LiveIn(int smpr, int chans, int *thedsp) {
  dsp = thedsp;
  sample_rate = smpr;
  channels = chans;

  num_samples=(int)((sample_rate*BUF_SIZE)/(sample_rate<<1));
  opt = (num_samples*channels);
  
  gotin = new short[opt];
  
  on = false;
}

LiveIn::~LiveIn() { 
  delete [ ] gotin;
};

int LiveIn::mix(int *mixpcm) {
  int res;

  res = get_audio();
  mixxx_stereo_44_novol(mixpcm,gotin,res);

  return(res);
}


int LiveIn::get_audio() {
  int res;

  res = read(*dsp,gotin,opt);

  return(res);
}
