/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>
#include <config.h>

#include <jutils.h>
#include "jmixer.h"
#include "audioproc.h"

/* process_buffer BUF_SIZE is:
   BUF_SIZE of 32bit ints *2channels *8resampling_space

   audio_buffer BUF_SIZE is:
   BUF_SIZE of 16bit short *2channels *8resampling_space
*/

Stream_mixer::Stream_mixer() {
  for(int i=0;i<MAX_CHANNELS;i++)
    xach[i] = NULL;

  /* here memset takes byte num */
  memset(process_buffer,0,PROCBUF_SIZE<<(SIZEOF_INT>>1));

  dsp = 0;
  max = 0;
  have_gui = false;

  dspout = false;
  linein = false;
  icecastout = false;
  fileout = false;
  quit = false;

  if(pthread_mutex_init (&_mutex,NULL) == -1)
    error("error initializing POSIX thread mutex");
  if(pthread_cond_init (&_cond, NULL) == -1)
    error("error initializing POSIX thread condtition"); 
}

Stream_mixer::~Stream_mixer() {
  int i;

  if(dsp>0)
    close_soundcard();
  
  if(icecastout)
    stop_icecast();

  for(i=0;i<MAX_CHANNELS;i++)
    /* delete_channel(i); */
    delete xach[i];
  
  if(pthread_mutex_destroy(&_mutex) == -1)
    error("error destroying POSIX thread mutex");
  if(pthread_cond_destroy(&_cond) == -1)
    error("error destroying POSIX thread condition");
}

void Stream_mixer::register_gui(Amuse *reg_gui) { 
  gui = reg_gui; 
  have_gui = true;
#ifdef HAVE_LAME
  lame.version();
  gui->status_lame->value(lame.ver);
#endif
  if(dsp<0) {
    gui->lineout_button->clear();
    gui->linein_button->clear();
  }
}

bool Stream_mixer::open_soundcard()
{
    int format,tstereo,speed,caps,val;
    
    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
      cerr << "[!]";
      perror("Error opening /dev/dsp: ");
      return(false);
    } else
      notice("Found soundcard on /dev/dsp");
    
    /* BUFFER FRAGMENTATION */
    val = (FRAGCOUNT << 16) | FRAGSIZE;

    if(ioctl(dsp,SNDCTL_DSP_SETFRAGMENT,&val)==-1)
      error("failed to set dsp buffers:");

    /* or LE */
    format = AFMT_S16_LE;

    tstereo = 1; /* only stereo _DSP_ in/out */
    speed = SAMPLE_RATE; /* 44100hz mixing */

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      act("full duplex supported. good");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      act("full duplex is not supported");
      fullduplex = false;
    }

    /* FORMAT */
    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      error("failed to set data format");

    /* CHANNELS */
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      error("something went wrong with the stereo setting");

    /* SAMPLERATE */
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      error("speed setting failed");

    /* GET FRAG SIZE BACK */
    if(ioctl(dsp,SNDCTL_DSP_GETBLKSIZE,&val)==-1)
      error("get block size failed to return");

    act("mixing 16bit %dHz stereo",speed);

    dspout = true;

    livein = new LiveIn(speed, tstereo+1, &dsp);

    return(true);
} /* open_soundcard */

bool Stream_mixer::set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps, char *name, char *url, char *desc, int login) {

  lock();

  if(!ice.set_server(ip,port,pass,mountpoint,bps,login)) {
    error("Can't resolve %s",ip);
    unlock();
    return(false);
  }

  ice.set_meta(name, url, desc);
  
  if((!lame.initialized)||(lame.bps!=bps)) {
    error("Ups! lame found uninitialized");
    unlock();
    return(false);
  }

  /* ::::: we are connected */
  if(ice.start()) {
    notice("estabilished %ubps stream on http://%s:%u/%s",ice.bps,ice.ip,ice.port,ice.mountpoint);
    icecastout = true;
  }
  /* ::::: connection failed */
  else icecastout = false;
  
  unlock();
  return(icecastout);
}

void Stream_mixer::set_lame(int bps, int freq, int mode, int quality, int lowpass, int highpass) {

  /* mode 2 (double channel) is unsupported */
  mode = (mode==2) ? 0 : mode;

  lock();
  lame.init(bps, freq, mode, quality, lowpass, highpass);
  unlock();

  char *modestr;
  switch(mode) {
  case 0:
    modestr = strdup("stereo");
    break;
  case 1:
    modestr = strdup("jstereo");
    break;
  case 3:
    modestr = strdup("mono");
    break;
  default:
    modestr = strdup(" ");
    break;
  }

  notice("lame encoding at %uKbit/s %s (q%u)",bps,modestr,quality);
  free(modestr);
}

void Stream_mixer::stop_icecast() {
  lock();
  ice.stop();
  icecastout = false;
  unlock();
  notice("interrupted streaming to server");
}

int Stream_mixer::set_mp3file(char *file, int bps) {
  lock();
  /*
  if((!lame.initialized)||(lame.bps!=bps))
  lame.init(bps);
  */
  fileout = mp3file.openmp3(file);

  unlock();
  notice("dumping to %s (%uKbit/s)",file,bps);
  return(fileout);
}

void Stream_mixer::stop_mp3file() {
  lock();
  mp3file.closemp3();
  fileout = false;
  unlock();
}

void Stream_mixer::cafudda()
{
  int i;
  float res;

  /* here memset takes byte num
     max *2 (16bit) *2 (stereo) */
  memset(process_buffer,0,max<<2);

  max = 0;

  if(quit) return;


  lock();

  for(i=0;i<MAX_CHANNELS;i++) {
    if(xach[i] != NULL) {
      if(xach[i]->on) {
	res = xach[i]->mix(process_buffer);
	max = (max<xach[i]->in_smp) ? xach[i]->in_smp : max;
	updchan(i,res);
	xach[i]->signal();
      } /* if(xach[i].on) */
    } /* if(xach[i] != NULL) */
  } /* for(i=0;i<MAX_CHANNELS;i++) */
  
  if(linein) {
    int ires;
    ires = livein->mix(process_buffer);
    max = (max<ires) ? ires : max;
  }
  
  /* here: max = number of 32bit samples in process_buffer
     number of single 16bit stereo samples (max>>1)
  */

  if(max>0) {
    /* we got some sound to process */

    /* CLIPPING
       this brings it back to a 16bit resolution
       i'm using a simple as effective alghoritm (audioproc.cpp) */
    clip_audio(max);

    /* ENCODE AUDIO
       here lame needs the number of BYTES
       so max*channels>>1 (16bit)
       (we mix in stereo domain so channels are always 2) */
    if(fileout||icecastout) {
      lame.encode(audio_buffer,max>>2);
    }
    
    /* WRITE 2 MP3 FILE */
    if(fileout) {
      mp3file.writemp3(lame.buffer,lame.encoded);
    }

    /* SEND 2 ICECAST */
    if(icecastout) {
      ice.send(lame.buffer,lame.encoded);
      if(have_gui)
	if(gui->levels->showit)
	  gui->levels->set_bitrate( lame.get_bitrate() );
    } else {
      if(have_gui)
	if(gui->levels->showit)
	  gui->levels->set_bitrate( 0 );
    }      
	
    unlock();

    /* WRITE 2 DSP */
    if(dspout) {
      /* write out interleaved stereo 16bit pcm 
	 dsp takes number of samples, the format
	 is being setted with ioctls in initialization
      */
      write(dsp,audio_buffer,max);
    }

    /* compute and draw levels
       notice the gui to refresh */
    if(have_gui)
      if(gui->levels->showit)
	gui->levels->vumeter->value( get_level(max) );

  } else {
    /* there was no sound coming in from any channel */

    if(have_gui) {
      if(gui->levels->showit) {
	gui->levels->vumeter->value( gui->levels->vumeter->minimum() );
	gui->levels->set_bitrate( gui->levels->bitrate->minimum() );
      }
    }

    wait();
    unlock();
  }
  

}

bool Stream_mixer::create_channel(int ch) {
  lock();

  /* paranoia */
  if(xach[ch]!=NULL)
    delete_channel(ch);

  /* by default we create a mpeg channel */
  xach[ch] = new MpegChannel();

  unlock();
  
  if(xach[ch]==NULL)
    return(false);

  return(true);
}

void Stream_mixer::delete_channel(int ch) { 
  lock();
  if(xach[ch]->on) xach[ch]->stop();
  if(xach[ch]->opened) xach[ch]->clean();
  if(xach[ch]->running) {
    /* be sure it quitted */
    xach[ch]->signal();
    xach[ch]->lock(); xach[ch]->unlock();
  }

  delete xach[ch];
  xach[ch] = NULL;

  unlock();
}

void Stream_mixer::pause_channel(int ch) {
  /* here i don't lock - c'mon, boolean _is_ atomical */
  if(xach[ch]->opened) {
    xach[ch]->on = !xach[ch]->on;
  }
}

/* set_channel returns:
   0 - error
   1 - bitstream opened (seekable)
   2 - bitstream opened (non seekable)
*/
int Stream_mixer::set_channel(int ch, int playlist_pos) {
  int res;
  char *sel = playlist[ch].song(playlist_pos);

  if(!sel) return(0);

  lock();

  if(strncasecmp(sel,".ogg",4)==0) {
#ifdef HAVE_VORBIS
    /* take care to delete the old channel if present*/
    if(xach[ch]!=NULL) {
      if(xach[ch]->type != OGGCHAN) {
	delete xach[ch];
	/* better to wait a bit after deleting */
	jsleep(0,100);
	xach[ch] = new OggChannel();
      }
    }
#else
    warning("can't open file: OggVorbis support not compiled.");
    return(false);
#endif
  } else if(strncasecmp(sel,".mp3",4)==0) {
    /* take care to delete the old channel if present*/
    if(xach[ch]!=NULL) {
      if(xach[ch]->type != MP3CHAN) {
	delete xach[ch];
	jsleep(0,100);
	xach[ch] = new MpegChannel();
      }
    }
  }

  unlock();

  /* ok, we have the channel here, let's load the bitstream */
  res = xach[ch]->set(playlist[ch].song(playlist_pos));
  if(!res)
    error("can't play %s", playlist[ch].song(playlist_pos));
  else {
    playlist[ch].select(playlist_pos);
    switch(xach[ch]->playmode) {
    case 0: /* LOOP :::::::::::::::::::::::: */
    case 1: /* CONT :::::::::::::::::::::::: */
      play_channel(ch);
      break;
    default: /* PLAY :::::::::::::::::::::::: */
      break;
    }
    
    switch(xach[ch]->type) {
    case MP3CHAN:
      notice("Mpeg on chan[%u] %ukbit %ukhz %s"
	     , ch, xach[ch]->bitrate, xach[ch]->samplerate,
	     (xach[ch]->channels==1) ? "mono" : "stereo" );
      break;
    case OGGCHAN:
      notice("OggVorbis on chan[%u] %ukhz %s"
	     , ch, xach[ch]->samplerate,
	     (xach[ch]->channels==1) ? "mono" : "stereo" );
      break;
    }
  }
  return(res);
}

void Stream_mixer::set_all_volumes(float *vol) {
  int ch;
  lock();
  for(ch=0;ch<MAX_CHANNELS;ch++) {
    if(xach[ch]!=NULL)
      xach[ch]->volume = vol[ch];
  }
  unlock();
}

void Stream_mixer::set_volume(int ch, float vol) {
  lock();
  xach[ch]->volume = vol;
  unlock();
}

void Stream_mixer::set_speed(int ch, int speed) {
  lock();
  xach[ch]->speed = speed;
  unlock();
}

bool Stream_mixer::play_channel(int ch) {
  bool res = false;
  if(xach[ch]->running && !xach[ch]->on) {
    lock();
    if(! (res = xach[ch]->play()) )
      error("can't play channel %u",ch);
    unlock();
  }
  return(res);
}
  
bool Stream_mixer::stop_channel(int ch) {
  bool res = false;
  if(xach[ch]->running) {
    lock();
    res = xach[ch]->stop();
    unlock();
  }
  return(res);
}

bool Stream_mixer::set_position(float pos, int ch) {
  bool res = false;
  if(xach[ch]->seekable && xach[ch]->running) {
    lock();
    res = xach[ch]->pos(pos);
    unlock();
    if(!res) error("can't seek bitstream on channel [%u]",ch);
  }
  return(res);
}

bool Stream_mixer::set_live(bool stat) {
  if(dsp<1) {
    warning("ignoring live-in: soundcard not initialized");
    return(false);
  }
  if(!((dspout)&&(!fullduplex)&&(stat))) {
    lock();
    livein->on = linein = stat;
    unlock();
  }
  return(livein->on);
}

bool Stream_mixer::set_lineout(bool stat) {
  if(dsp<1) {
    warning("ignoring dspout: soundcard not initialized");
    return(false);
  }
  if(!((livein->on)&&(!fullduplex)&&(stat))) {
    lock();
    dspout = stat;
    unlock();
  }
  return(dspout&stat);
}

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

bool Stream_mixer::set_playlist(char *file, int ch) {
  bool res;
  lock();
  res = playlist[ch].load(file);
  unlock();
  return(res);
}

void Stream_mixer::set_playmode(int ch, int mode) {
  
  switch(mode) {
  case 0:
    xach[ch]->playmode = xach[ch]->PLAY;
    break;
  case 1:
    xach[ch]->playmode = xach[ch]->LOOP;
    break;
  case 2:
    xach[ch]->playmode = xach[ch]->CONT;
    break;
  }
}

bool Stream_mixer::add_to_playlist(char *file, int ch) {
  lock();
  playlist[ch].add(file);
  unlock();
  return(true);
}

/* updchan takes care that any action is taken (channel ends etc.)
   and updates the gui if registered.
   this is called only by cafudda while the mixer is locked, so no need to lock
*/
void Stream_mixer::updchan(int ch, float res) {

  if(res==3.0) {
    warning("uops! channel [%u] out of sync",ch);
    return;
  } else if(res==2.0) {

    if(xach[ch]->rewinding) return;

    unlock();
    func("Stream_mixer::updchan(%u,%f)",ch,res);
    switch(xach[ch]->playmode) {
    case 0: /* LOOP :::::::::::::::::::::::: */     
      xach[ch]->set(playlist[ch].selected->name);
      xach[ch]->play();
      break;
    case 1: /* CONT :::::::::::::::::::::::: */
      if(!playlist[ch].selected->next)
	playlist[ch].select(1);
      else
	playlist[ch].select(playlist[ch].selected->next);
      
      xach[ch]->set(playlist[ch].selected->name);
      xach[ch]->play();
      break;
    case 2: /* PLAY :::::::::::::::::::::::: */
      xach[ch]->set(playlist[ch].selected->name);
      break;
    }

    xach[ch]->rewinding = true;

    lock();

  } else if(res<=1.0) xach[ch]->rewinding = false;
  
  if(have_gui) {
    char temp[64];
    sprintf(temp,"%02u:%02u:%02u",xach[ch]->time.h,xach[ch]->time.m,xach[ch]->time.s);
    gui->chan[ch]->upd(res, xach[ch]->playmode, temp);
  }
}

void Stream_mixer::clip_audio(int chunks) {
  int c;
  if(chunks==0) return;
  for(c=0;c<chunks;c++) {
    if(process_buffer[c]>32767) {
      audio_buffer[c]=32767;
    } else if(process_buffer[c]<-32767) {
      audio_buffer[c]=-32767;
    } else {
      audio_buffer[c]=(short)process_buffer[c];
    }
  }
}

int Stream_mixer::get_level(int chunks) {
  int c;
  int peak = 0;
  int low = 32767;
  //  unsigned long int round;

  for(c=0;c<chunks-2;c++) {
    if(audio_buffer[c]>peak)
      peak = (int) audio_buffer[c];
    if(audio_buffer[c]<low)
      low = (int) audio_buffer[c];
  }

  /*
  round = (unsigned long int) peak;
  round = (round<<7)>>14;
  return((unsigned int)round);
  */
  return((peak+low)>>1);
}
