/* MuSE - Multiple Streaming Engine
 * Copyright (C) 2000-2002 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * here command line is parsed
 * synchronous threads are launched depending on mode selection
 * signals are trapped for nice&clean exiting
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>

#include <jutils.h>
#include "config.h"
#include "generic.h"
#include "jmixer.h"
#include "gui_muse.h"
#include "gui_callbacks.h"

/* command line stuff */

char *version =
"%s version %s [ http://muse.dyne.org ]";

char *authors =
"(C) 2000-2002 jaromil & august black";

char *help =
"Usage: muse [generic options] [-c [encoder options] [stream options]]\n"
":: generic options:\n"
"-h --help         this help\n"
"-v --version      version information\n"
"-d --debug [1-3]  debug verbosity level       - default 1\n"
"-o --nodsp        disable souncard support\n"
"-c --cli          command line input (no GUI)\n"
":: encoder options:\n"
"-b --bitrate      encoding bitrate in Kbit/s  - default 24\n"
"-q --quality      encoding quality (1-9)      - default 4\n"
"-f --filedump     dump stream to file\n"
"-i --live         run listening on live input\n"
"-l --playlist     run cycling thru a playlist\n"
":: stream options:\n"
"-s --server       stream to server\n"
"-p --port         server port                 - default 8000\n"
"-m --mountpoint   mounpoint on server         - default live\n"
"-x --pass         encoder password on server\n"
"-n --name         name of the stream\n"
"\n";

static const struct option long_options[] = {
  { "help", no_argument, NULL, 'h' },
  { "version", no_argument, NULL, 'v' },
  { "debug", required_argument, NULL, 'd' },
  { "nodsp", no_argument, NULL, 'o' },
  { "bitrate", required_argument, NULL, 'b' },
  { "quality", required_argument, NULL, 'q' },
  { "filedump", required_argument, NULL, 'f' },
  { "cli", no_argument, NULL, 'c' },
  { "live", no_argument, NULL, 'i' },
  { "playlist", required_argument, NULL, 'l' },
  { "server", required_argument, NULL, 's' },
  { "port", required_argument, NULL, 'p' },
  { "mountpoint", required_argument, NULL, 'm' },
  { "pass",required_argument, NULL, 'x' },
  { "name",required_argument, NULL, 'n' },
  { 0, 0, 0, 0 }
};

char *short_options = "hvd:ob:q:f:cil:s:p:m:x:n:";

/* misc settings */
int debug = 1;
bool daemon_mode = false;
char *queue_file = NULL;

int brate = 24;
int quality = 4;
char *filedump = NULL;
bool nogui = false;
bool live = false;
char *playlist = NULL;
char *server = NULL;
int port = 8000;
char *mountpoint = NULL;
char *pass = NULL;
char *name = NULL;
bool nodsp = false;

/* declare the sighandler */
void quitproc (int Sig);
volatile sig_atomic_t userbreak;
/* ---------------------- */

void take_args(int argc, char **argv) {
  int res;

  notice(version,PACKAGE,VERSION,PACKAGE);
  act(authors);
  set_debug(3);
  
  do {
    res = getopt_long(argc, argv, short_options, long_options, NULL);

    switch(res) {
    case 'h':
      fprintf(stderr,"%s",help);
      exit(0);
      break;
    case 'v':
      exit(0);
      break;
    case 'd':
      debug = atoi(optarg);
      break;
    case 'o':
      nodsp = true;
      break;
    case 'b':
      brate = atoi(optarg);
      break;
    case 'q':
      quality = atoi(optarg);
      break;
    case 'f':
      filedump = strdup(optarg);
      break;
    case 'c':
      nogui = true;
      break;
    case 'i':
      live = true;
      break;
    case 'l':
      playlist = strdup(optarg);
      break;
    case 's':
      server = strdup(optarg);
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 'm':
      mountpoint = strdup(optarg);
      break;
    case 'x':
      pass = strdup(optarg);
      break;
    case 'n':
      name = strdup(optarg);
      break;
    default:
      break;
    }
  } while(res > 0);

}

Stream_mixer *mix;

int main(int argc, char **argv) {
  mix = NULL;
  Amuse *gui = NULL;

  take_args(argc, argv);

  /* register signal traps */
  if (signal (SIGINT, quitproc) == SIG_ERR) {
    error ("Couldn't install SIGINT handler"); exit (1);
  }
  if (signal (SIGQUIT, quitproc) == SIG_ERR) {
    error ("Couldn't install SIGQUIT handler"); exit (1);
  }
  if (signal (SIGTERM, quitproc) == SIG_ERR) {
    error ("Couldn't install SIGTERM handler"); exit (1);
  }

  set_debug(debug);
  act("starting engine...");
  
  mix = new Stream_mixer();

  if(!nodsp)
    if(!mix->open_soundcard()) {
      warning("soundcard not found");
      act("line-in and speaker out deactivated");
    }
    
  if(!nogui) { /* we want a GUI */
    notice("spawning GUI");
    gui = new Amuse(310, 70, PACKAGE);
    init_callbacks(mix,gui);
    gui->start();
  } else { /* run from CLI */
    /* lame guesses for best frequency
       hardcoded mono channel (mode 3)
       lame guesses for best low/highpass filtering
    */
    mix->set_lineout(false);
    mix->set_lame(brate,0,3,quality,0,0);
    if(filedump) mix->set_mp3file(filedump,brate);
    if( mix->set_live(live) ) notice("recording from live input");
    if(playlist) {
      notice("looping playlist %s",playlist);
      if(!mix->create_channel(0))
	error("strange! i got problems creating a channel");
      else if(!mix->set_playlist(playlist,0))
	error("hey! i can't load %s as a playlist",playlist);
      else if(!mix->set_channel(0,1))
	error("jeez! can't set %s on channel",mix->playlist[0].selected->name);
      mix->set_playmode(0,2);
      mix->play_channel(0);

    }
    if(server)
      mix->set_icecast(server,port,pass,mountpoint,brate,name,NULL,NULL,0);
  }
  
  while(!mix->quit)
    mix->cafudda();
  
  notice("quitting MuSE");
  jsleep(1,0);

  if(mix) delete mix;
  if(gui) delete gui;
  act("cya!");
  
  exit(0);
}

/* signal handling */
void quitproc (int Sig) {
  notice("received signal %u on process %u",Sig,getpid());
  mix->quit = true;
}
